-- 
-- This SQL Script creates all TriggerMan database objects. 
--

CREATE SCHEMA IF NOT EXISTS triggerman; 

CREATE TABLE IF NOT EXISTS triggerman.tman_options (
  ID int(11) NOT NULL AUTO_INCREMENT,
  schema_name varchar(64) NOT NULL,
  audit_table_prefix varchar(64) NOT NULL DEFAULT '',
  audit_table_suffix varchar(64) NOT NULL DEFAULT '_audit_tman',
  trigger_name_format varchar(64) NOT NULL DEFAULT '%TBL%_%TIMING%_%EVENT%_tman',
  before_string varchar(16) NOT NULL DEFAULT 'bef',
  after_string varchar(16) NOT NULL DEFAULT 'aft',
  delete_string varchar(16) NOT NULL DEFAULT 'del',
  insert_string varchar(16) NOT NULL DEFAULT 'ins',
  update_string varchar(16) NOT NULL DEFAULT 'upd',
  audit_schema_name varchar(64) NOT NULL,
  tmp_audit_table_prefix varchar(64) NOT NULL DEFAULT 'tmp_',
  auto_BD bit(1) NOT NULL DEFAULT b'0',
  auto_BI bit(1) NOT NULL DEFAULT b'0',
  auto_BU bit(1) NOT NULL DEFAULT b'0',
  auto_AD bit(1) NOT NULL DEFAULT b'1',
  auto_AI bit(1) NOT NULL DEFAULT b'1',
  auto_AU bit(1) NOT NULL DEFAULT b'1',
  auto_save_restore bit(1) NOT NULL DEFAULT b'1', 
  auto_rebuild_audit bit(1) NOT NULL DEFAULT b'1', 
  PRIMARY KEY (ID),
  UNIQUE KEY ID_UNIQUE (ID), 
  UNIQUE KEY SCHEMA_UNIQUE (SCHEMA_NAME));

DELIMITER $$ 


CREATE FUNCTION triggerman.getColumnName( tableName VARCHAR(64), schemaName VARCHAR(64) ) RETURNS VARCHAR(2000)
BEGIN

  DECLARE done INT DEFAULT FALSE;
  DECLARE first BOOLEAN DEFAULT TRUE; 
  DECLARE colName VARCHAR(64) DEFAULT ''; 
  DECLARE colStr VARCHAR(2000) DEFAULT ''; 

  DECLARE cur1 CURSOR FOR 
	SELECT column_name 
	FROM information_schema.columns
	WHERE table_name = tableName AND table_schema = schemaName; 


  DECLARE CONTINUE HANDLER FOR NOT FOUND SET done = TRUE;

-- Create string of column names separated by commas, with no comma after the last column definition...
  OPEN cur1;

  read_loop: LOOP
    FETCH cur1 INTO colName;

    IF done THEN
      LEAVE read_loop;
    END IF;

	IF first THEN 
		SET first = FALSE; 
		SET colStr = colName; 
    ELSE 
		SET colStr = CONCAT( colStr, ', ', colName ); 
	END IF; 

  END LOOP;

  CLOSE cur1;

  RETURN colStr; 

END$$


CREATE FUNCTION triggerman.getColumnNamePrepend( tableName VARCHAR(64), schemaName VARCHAR(64), prependStr VARCHAR(64) ) RETURNS VARCHAR(2000)
BEGIN

  DECLARE done INT DEFAULT FALSE;
  DECLARE first BOOLEAN DEFAULT TRUE; 
  DECLARE colName VARCHAR(64) DEFAULT ''; 
  DECLARE colStr VARCHAR(2000) DEFAULT ''; 

  DECLARE cur1 CURSOR FOR 
	SELECT CONCAT( prependStr, column_name ) 
	FROM information_schema.columns
	WHERE table_name = tableName AND table_schema = schemaName; 


  DECLARE CONTINUE HANDLER FOR NOT FOUND SET done = TRUE;

-- Create string of column names separated by commas, with no comma after the last column definition...
  OPEN cur1;

  read_loop: LOOP
    FETCH cur1 INTO colName;

    IF done THEN
      LEAVE read_loop;
    END IF;

	IF first THEN 
		SET first = FALSE; 
		SET colStr = colName; 
    ELSE 
		SET colStr = CONCAT( colStr, ', ', colName ); 
	END IF; 

  END LOOP;

  CLOSE cur1;

  RETURN colStr; 

END$$


CREATE FUNCTION triggerman.getColumnNameIntersect( tableName1 VARCHAR(64), tableName2 VARCHAR(64), 
							schemaName VARCHAR(64) ) RETURNS VARCHAR(2000)
BEGIN

  DECLARE done INT DEFAULT FALSE;
  DECLARE first BOOLEAN DEFAULT TRUE; 
  DECLARE colName VARCHAR(64) DEFAULT ''; 
  DECLARE colStr VARCHAR(2000) DEFAULT ''; 

  DECLARE cur1 CURSOR FOR 
	SELECT c1.column_name 
	FROM information_schema.columns c1
	INNER JOIN information_schema.columns c2 ON c2.column_name = c1.column_name
	WHERE c1.table_name = tableName1 AND c1.table_schema = schemaName AND
		  c2.table_name = tableName2 AND c2.table_schema = schemaName; 


  DECLARE CONTINUE HANDLER FOR NOT FOUND SET done = TRUE;

-- Create string of column names separated by commas, with no comma after the last column definition...
  OPEN cur1;

  read_loop: LOOP
    FETCH cur1 INTO colName;

    IF done THEN
      LEAVE read_loop;
    END IF;

	IF first THEN 
		SET first = FALSE; 
		SET colStr = colName; 
    ELSE 
		SET colStr = CONCAT( colStr, ', ', colName ); 
	END IF; 

  END LOOP;

  CLOSE cur1;

  RETURN colStr; 

END$$


CREATE FUNCTION triggerman.getColumnDef( tableName VARCHAR(64), schemaName VARCHAR(64) ) RETURNS VARCHAR(2000)
BEGIN

  DECLARE done INT DEFAULT FALSE;
  DECLARE first BOOLEAN DEFAULT TRUE; 
  DECLARE colDef VARCHAR(200) DEFAULT ''; 
  DECLARE colStr VARCHAR(2000) DEFAULT ''; 

  DECLARE cur1 CURSOR FOR 
	SELECT CONCAT( column_name, ' ', column_type, ' ', IF( is_nullable = 'yes', 'NULL', 'NOT NULL' ), ' ',	
	IF( column_default IS NOT NULL, CONCAT( 'DEFAULT ', column_default ), '') ) ColumnDef  -- NoAutoInc.., ' ',extra, ',' ) ColumnDef
	FROM information_schema.columns
	WHERE table_name = tableName AND table_schema = schemaName; 


  DECLARE CONTINUE HANDLER FOR NOT FOUND SET done = TRUE;

-- Create string of column definitions separated by commas, with no comma after the last column definition...
  OPEN cur1;

  read_loop: LOOP
    FETCH cur1 INTO colDef;

    IF done THEN
      LEAVE read_loop;
    END IF;

	IF first THEN 
		SET first = FALSE; 
		SET colStr = colDef; 
    ELSE 
		SET colStr = CONCAT( colStr, ', ', colDef ); 
	END IF; 

  END LOOP;

  CLOSE cur1;

  RETURN colStr; 

END$$


CREATE PROCEDURE triggerman.getAuditData( IN schemaName VARCHAR(64), IN auditSchemaName VARCHAR(64) )
BEGIN

DECLARE auditPrefix VARCHAR(64); 
DECLARE auditSuffix VARCHAR(64); 


-- Get Audit Table prefix and suffix for this schema. 
	SELECT	audit_table_prefix, audit_table_suffix
	INTO 	auditPrefix, auditSuffix
	FROM	triggerman.tman_options
	WHERE   schema_name = schemaName; 

-- If audit schema is blank, use table schema
	IF auditSchemaName = '' OR auditSchemaName IS NULL THEN
		SET auditSchemaName = schemaName; 
	END IF; 	

-- Get table, audit table and trigger data
SELECT t.table_name 'Table', IFNULL(aud.table_name, '') AuditTable, 
					 IFNULL(bd.trigger_name,'') BDTrigger, IFNULL(bi.trigger_name,'') BITrigger, IFNULL(bu.trigger_name,'') BUTrigger, 
					 IFNULL(ad.trigger_name,'') ADTrigger, IFNULL(ai.trigger_name,'') AITrigger, IFNULL(au.trigger_name,'') AUTrigger
FROM information_schema.tables t
LEFT OUTER JOIN information_schema.tables aud ON aud.table_name = CONCAT( auditPrefix, t.table_name, auditSuffix ) AND
								aud.table_schema = auditSchemaName  -- Audit Table schema can be different than base table
LEFT OUTER JOIN information_schema.triggers bd ON t.table_name = bd.event_object_table AND t.table_schema = bd.event_object_schema AND 
						       bd.action_timing = 'BEFORE' AND bd.event_manipulation = 'DELETE'
LEFT OUTER JOIN information_schema.triggers bi ON t.table_name = bi.event_object_table AND t.table_schema = bi.event_object_schema AND 
						       bi.action_timing = 'BEFORE' AND bi.event_manipulation = 'INSERT'
LEFT OUTER JOIN information_schema.triggers bu ON t.table_name = bu.event_object_table AND t.table_schema = bu.event_object_schema AND
							   bu.action_timing = 'BEFORE' AND bu.event_manipulation = 'UPDATE'
LEFT OUTER JOIN information_schema.triggers ad ON t.table_name = ad.event_object_table AND t.table_schema = ad.event_object_schema AND 
						       ad.action_timing = 'AFTER' AND ad.event_manipulation = 'DELETE'
LEFT OUTER JOIN information_schema.triggers ai ON t.table_name = ai.event_object_table AND t.table_schema = ai.event_object_schema AND 
						       ai.action_timing = 'AFTER' AND ai.event_manipulation = 'INSERT'
LEFT OUTER JOIN information_schema.triggers au ON t.table_name = au.event_object_table AND t.table_schema = au.event_object_schema AND
							   au.action_timing = 'AFTER' AND au.event_manipulation = 'UPDATE'
WHERE t.table_schema = schemaName AND t.table_type = 'BASE TABLE' AND 
	NOT( INSTR(t.table_name, auditPrefix) = 1 AND INSTR(t.table_name, auditSuffix) = 
												LENGTH(t.table_name) - LENGTH(auditSuffix) + 1); 

END$$


CREATE PROCEDURE triggerman.getTmanInfo( IN schemaName VARCHAR(64) )
BEGIN

-- Get table, audit table and trigger data
	SELECT audit_table_prefix, audit_table_suffix, trigger_name_format, before_string, after_string,
		   delete_string, insert_string, update_string, audit_schema_name, tmp_audit_table_prefix,  
		   auto_BD, auto_BI, auto_BU, auto_AD, auto_AI, auto_AU, auto_save_restore, auto_rebuild_audit
FROM triggerman.tman_options 
	WHERE schema_name = schemaName; 

END$$


CREATE PROCEDURE triggerman.saveTmanInfo( IN schemaName VARCHAR(64), IN auditTablePrefix VARCHAR(64), IN auditTableSuffix VARCHAR(64),
							IN triggerNameFormat VARCHAR(64), beforeString VARCHAR(16), afterString VARCHAR(16), 
							IN deleteString VARCHAR(16), insertString VARCHAR(16), IN updateString VARCHAR(16), 
							IN auditSchemaName VARCHAR(64), IN tmpAuditTablePrefix VARCHAR(64), 
							IN autoBD boolean, IN autoBI boolean, IN autoBU boolean, IN autoAD boolean, 
							IN autoAI boolean, IN autoAU boolean, IN autoSaveRestore boolean, 
							IN autoRebuildAudit boolean )
BEGIN

DECLARE schemaCount INT; 

-- Is schema name already in Options table? 
	SELECT  COUNT(ID)
	INTO 	schemaCount
	FROM	triggerman.tman_options
	WHERE   schema_name = schemaName; 

-- UPDATE or INSERT Options for the specified schema.  
	IF schemaCount = 0 THEN 

		INSERT INTO triggerman.tman_options
			( schema_name, audit_table_prefix, audit_table_suffix, trigger_name_format, before_string, after_string,
			  delete_string, insert_string, update_string, audit_schema_name, tmp_audit_table_prefix,
			  auto_BD, auto_BI, auto_BU, auto_AD, auto_AI, auto_AU, auto_save_restore, auto_rebuild_audit ) 
			VALUES ( schemaName, auditTablePrefix, auditTableSuffix, triggerNameFormat, beforeString, afterString, 
			  deleteString, insertString, updateString, auditSchemaName, tmpAuditTablePrefix,  
			  autoBD, autoBI, autoBU, autoAD, autoAI, autoAU, autoSaveRestore, autoRebuildAudit ); 

	ELSE

		UPDATE triggerman.tman_options
		SET audit_table_prefix = auditTablePrefix, audit_table_suffix = auditTableSuffix,
			trigger_name_format = triggerNameFormat, before_string = beforeString, after_string = afterString, 
			delete_string = deleteString, insert_string = insertString, update_string = updateString, 
		    audit_schema_name = auditSchemaName, tmp_audit_table_prefix = tmpAuditTablePrefix, 
			auto_BD = autoBD, auto_BI = autoBI, auto_BU = autoBU, auto_AD = autoAD, auto_AI = autoAI, auto_AU = autoAU, 
			auto_save_restore = autoSaveRestore, auto_rebuild_audit = autoRebuildAudit
		WHERE schema_name = schemaName; 

	END IF; 

END$$


CREATE PROCEDURE triggerman.setTmanDefault( IN schemaName VARCHAR(64) )
BEGIN

-- Set new entry in TMan_Options, uses defaults. 

	INSERT INTO triggerman.tman_options ( schema_name, audit_schema_name ) VALUES( schemaName, schemaName ); 
	COMMIT; 

END$$


CREATE FUNCTION triggerman.constructTriggerName(tableName VARCHAR(64), schemaName VARCHAR(64),  
				triggerTime VARCHAR(6), triggerEvent VARCHAR(6)) 
								  RETURNS VARCHAR(1000) 
BEGIN
 
	DECLARE output_string VARCHAR(64); 

	DECLARE before_string VARCHAR(16); 
	DECLARE after_string VARCHAR(16);
	DECLARE delete_string VARCHAR(16); 
	DECLARE insert_string VARCHAR(16);
	DECLARE update_string VARCHAR(16); 
	DECLARE trigger_name_format VARCHAR(64); 


	-- The trigger_name_format string can contain substitution variables: 
	--
	-- 	%TBL% - Substitute table name
	-- 	%TIMING% - Substitute timing: BEFORE or AFTER
	-- 	%EVENT% - Substitute event: INSERT, DELETE or UPDATE
	--
	-- Similarly, strings can be defined for the timing and event types. 

	SELECT	top.before_string, top.after_string, top.delete_string, top.insert_string, top.update_string, top.trigger_name_format	
	INTO	before_string, after_string, delete_string, insert_string, update_string, trigger_name_format
	FROM	triggerman.tman_options top
	WHERE	top.schema_name = schemaName; 


	-- If format string is NULL, just return a blank string. 
	IF trigger_name_format IS NULL OR trigger_name_format = '' THEN
		RETURN ''; 
	END IF; 

	-- If none of the substitution variables appear in the format string, 
	-- just return the format string ( Not Useful, static trigger name...)
	IF LOCATE(trigger_name_format, '%TBL%') = 0 AND 
	   LOCATE(trigger_name_format, '%TIMING%') = 0 AND 
	   LOCATE(trigger_name_format, '%EVENT%') = 0 THEN
		RETURN trigger_name_format; 
	END IF; 

	-- Perform substitution
	SET output_string = trigger_name_format; 
	SET output_string = REPLACE( output_string, '%TBL%', LOWER(tableName));
	SET output_string = REPLACE( output_string, '%TIMING%', LOWER(triggerTime)); 
	SET output_string = REPLACE( output_string, '%EVENT%', LOWER(triggerEvent)); 

	RETURN output_string; 

END$$


CREATE FUNCTION triggerman.getAuditTblDef( tableName VARCHAR(64), auditTableName VARCHAR(64), 
				schemaName VARCHAR(64), auditSchemaName VARCHAR(64) ) RETURNS VARCHAR(2000)
BEGIN

  DECLARE colStr VARCHAR(1000) DEFAULT ''; 
  DECLARE tblStr VARCHAR(2000) DEFAULT ''; 

  DECLARE idStr VARCHAR(50) DEFAULT '( hist_id INT(11) NOT NULL AUTO_INCREMENT, '; 
  DECLARE timeStr VARCHAR(60) DEFAULT 'hist_time DATETIME NOT NULL, '; 
  DECLARE userStr VARCHAR(50) DEFAULT 'hist_user VARCHAR(16) NOT NULL, '; 
  DECLARE typeStr VARCHAR(50) DEFAULT 'hist_type  VARCHAR(1) NOT NULL, '; 
  DECLARE keyStr VARCHAR(60) DEFAULT 'PRIMARY KEY (hist_id), UNIQUE KEY ID_UNIQUE (hist_id)) ';

	SET colStr = CONCAT( triggerman.getColumnDef( tableName, schemaName ), ', ');  -- Add comma at end of column def string

	SET tblStr = CONCAT( 'CREATE TABLE IF NOT EXISTS ', auditSchemaName, '.', auditTableName, 
					idStr, timeStr, userStr, typeStr, colStr, keyStr ); 

	RETURN tblStr; 

END$$


CREATE FUNCTION triggerman.getTriggerDef( tableName VARCHAR(64), auditTableName VARCHAR(64), triggerName VARCHAR(64), schemaName VARCHAR(64), 
				auditSchemaName VARCHAR(64), triggerTime VARCHAR(6), triggerEvent VARCHAR(6) ) RETURNS VARCHAR(2000)
BEGIN

  DECLARE colStr VARCHAR(2000) DEFAULT ''; 
  DECLARE colStrPrepend VARCHAR(2000) DEFAULT ''; 
  DECLARE trgStr VARCHAR(4000) DEFAULT ''; 
  DECLARE trgName VARCHAR(64) DEFAULT '';
  DECLARE prependStr VARCHAR(4); 

  DECLARE histStr VARCHAR(60) DEFAULT 'hist_time, hist_user, hist_type, ';   -- Hist_ID not included, AutoInc...
  DECLARE valStr VARCHAR(60) DEFAULT  'NOW(), USER(), '; 

	-- Determine if we're using old or new values. 
	IF triggerEvent = 'DELETE' THEN 
		SET prependStr = 'OLD.'; 
	ELSE
		SET prependStr = 'NEW.'; 
	END IF; 

	-- Get column name strings, with and without prepend string. 
	SET colStr = triggerman.getColumnName( tableName, schemaName ); 
	SET colStrPrepend = triggerman.getColumnNamePrepend( tableName, schemaName, prependStr ); 

	-- Build CREATE TRIGGER statement. 
	SET trgStr = CONCAT( 'CREATE TRIGGER ', triggerName, ' ', triggerTime, ' ', triggerEvent, ' ON ', schemaName, '.',  tableName, ' FOR EACH ROW BEGIN ' ); 
	SET trgStr = CONCAT( trgStr, 'INSERT INTO ', auditSchemaName, '.', auditTableName, '( ', histStr, colStr, ' ) VALUES (', valStr, '''' ); 
	SET trgStr = CONCAT( trgStr, SUBSTR(triggerEvent,1,1), ''', ', colStrPrepend, ' ); END;' ); 

	RETURN trgStr; 

END$$


CREATE FUNCTION triggerman.restoreTmpTable( tableName VARCHAR(64), schemaName VARCHAR(64),
				tmpTblPrefix VARCHAR(64) ) RETURNS VARCHAR(2000)
BEGIN

	DECLARE colStr VARCHAR(1000); 
	DECLARE restoreStr VARCHAR(2000); 


	SET colStr = triggerman.getColumnNameIntersect( tableName, CONCAT(tmpTblPrefix, tableName), schemaName ); 

	SET restoreStr = CONCAT( 'INSERT INTO ', schemaName, '.', tableName, '( ', colStr, ' ) ' ); 
	SET restoreStr = CONCAT( restoreStr, 'SELECT ', colStr, ' FROM ', schemaName, '.', tmpTblPrefix, tableName ); 

	RETURN restoreStr; 

END$$


CREATE FUNCTION triggerman.compareAuditTbl( tableName VARCHAR(64), tableSchema VARCHAR(64), 
											auditTableName VARCHAR(64), auditTableSchema VARCHAR(64)) RETURNS boolean
BEGIN

	DECLARE colNum int DEFAULT 0; 
	DECLARE tabColNum int DEFAULT 0; 
	DECLARE auditColNum int DEFAULT 0; 
	DECLARE numAuditColumns int DEFAULT 4;  -- Number of Historic Audit Table columns

	-- If Audit Table DNE, return True, to indicate non-error...
	IF auditTableName IS NULL OR auditTableSchema IS NULL THEN 
		RETURN true; 
	END IF; 

	-- Get number of base table columns. 
	SELECT COUNT(c.column_name) 
	INTO   colNum
	FROM   information_schema.columns c
	WHERE  c.table_name = tableName AND 
		   c.table_schema = tableSchema; 

	-- Get number of matching audit table columns. 
	SELECT COUNT(tab.column_name) 
	INTO   tabColNum
		   --	tab.table_schema, tab.table_name, tab.column_name, tab.column_type, tab.is_nullable, tab.column_default, 
		   --	aud.table_schema, aud.table_name, aud.column_name, aud.column_type, aud.is_nullable, aud.column_default
	FROM   information_schema.columns tab, information_schema.columns aud
	WHERE  tab.table_name = tableName AND tab.table_schema = tableSchema AND
		   aud.table_name = auditTableName AND aud.table_schema = auditTableSchema AND
		   tab.column_name = aud.column_name AND 
		   tab.column_type = aud.column_type AND 
		   tab.is_nullable = aud.is_nullable AND 
		   IFNULL(tab.column_default,'NULL') = IFNULL(aud.column_default,'NULL'); 

	-- Get number of extra audit table columns
	SELECT 	COUNT(aud.column_name)
	INTO 	auditColNum
	FROM    information_schema.columns aud
	WHERE   aud.table_name = auditTableName AND aud.table_schema = auditTableSchema AND	
			aud.column_name NOT IN ( SELECT column_name 
									 FROM information_schema.columns 
									 WHERE table_name = tableName AND table_schema = tableSchema ); 


	IF colNum = tabColNum AND auditColNum = numAuditColumns THEN
		RETURN true; 
	ELSE
		RETURN false; 
	END IF; 

END$$


